"""PySerials base Exception class."""

from __future__ import annotations

from typing import TYPE_CHECKING as _TYPE_CHECKING
from functools import partial as _partial

from exceptionman import ReporterException as _ReporterException
import mdit as _mdit

if _TYPE_CHECKING:
    from mdit import Document


class PySerialsException(_ReporterException):
    """Base class for all exceptions raised by PySerials."""

    def __init__(self, report: Document):
        sphinx_config = {"html_title": "PySerials Error Report"}
        sphinx_target_config = _mdit.target.sphinx(
            renderer=_partial(
                _mdit.render.sphinx,
                config=_mdit.render.get_sphinx_config(sphinx_config)
            )
        )
        report.target_configs["sphinx"] = sphinx_target_config
        super().__init__(report=report)
        return



import jsonschema as _jsonschema
import referencing as _referencing
import referencing.jsonschema as _referencing_jsonschema
from referencing.exceptions import Unresolvable as _UnresolvableReferencingError
import pyserials.exception as _exception


def jsonschema(
    data: dict | list | str | int | float | bool,
    schema: dict,
    validator: type[_jsonschema.protocols.Validator] = _jsonschema.Draft202012Validator,
    registry: _referencing_jsonschema.SchemaRegistry | None = None,
    fill_defaults: bool = True,
    iter_errors: bool = False,
    raise_invalid_data: bool = True,
) -> list[_jsonschema.exceptions.ValidationError]:
    """
    Validate data against a JSON schema.

    Parameters
    ----------
    data : dict | list | str | int | float | bool
        The data to validate.
    schema : dict
        The schema to validate the data against.
    validator : jsonschema.protocols.Validator, default: jsonschema.Draft202012Validator
        The JSON schema validator to use.
    fill_defaults : bool, default: True
        Whether to fill in the data with default values from the schema,
        when they are not present.
    raise_invalid_data : bool, default: True
        Whether to raise an exception when the data is invalid against the schema.

    Returns
    -------
    valid : bool
        Whether the data is valid against the schema.

    Raises
    ------
    pyserials.exception.PySerialsInvalidSchemaError
        If the schema is invalid.
    pyserials.exception.PySerialsSchemaValidationError
        If the data is invalid against the schema and `raise_invalid_data` is `True`.
    """
    def _extend_with_default(validator_class: _jsonschema.protocols.Validator) -> _jsonschema.protocols.Validator:
        # https://python-jsonschema.readthedocs.io/en/stable/faq/#why-doesn-t-my-schema-s-default-property-set-the-default-on-my-instance

        validate_properties = validator_class.VALIDATORS["properties"]

        def set_defaults(validator, properties, instance, schema):
            if isinstance(instance, dict):  # The entire dict instance may be templated
                for property, subschema in properties.items():
                    if "default" in subschema:
                        instance.setdefault(property, subschema["default"])

            for error in validate_properties(
                validator,
                properties,
                instance,
                schema,
            ):
                yield error

        return _jsonschema.validators.extend(validator_class, {"properties": set_defaults})

    error_args = {"data": data, "schema": schema, "validator": validator, "registry": registry}
    validator = _extend_with_default(validator) if fill_defaults else validator
    try:
        validator_instance = validator(schema, registry=registry) if registry else validator(schema)
    except (
        _jsonschema.exceptions.SchemaError,
        _jsonschema.exceptions.UndefinedTypeCheck,
        _jsonschema.exceptions.UnknownType,
        _UnresolvableReferencingError,
    ) as e:
        raise _exception.validate.PySerialsInvalidJsonSchemaError(**error_args) from e
    if iter_errors:
        errors = list(validator_instance.iter_errors(data))
    else:
        try:
            validator_instance.validate(data)
            errors = []
        except (_jsonschema.exceptions.ValidationError, _jsonschema.exceptions.FormatError) as e:
            errors = [e]
    if raise_invalid_data and errors:
        error_args["validator"] = validator_instance
        raise _exception.validate.PySerialsJsonSchemaValidationError(**error_args, causes=errors)
    return errors



from __future__ import annotations as _annotations

from typing import TYPE_CHECKING as _TYPE_CHECKING
import re as _re
from functools import partial as _partial

import jsonpath_ng as _jsonpath
from jsonpath_ng import exceptions as _jsonpath_exceptions

import pyserials.exception as _exception

if _TYPE_CHECKING:
    from typing import Literal, Sequence, Any, Callable, Iterable
    UPDATE_OPTIONS = Literal["skip", "write", "raise"] | Callable[
        [tuple[Any] | tuple[Any, Any]], None | tuple[Any, str]
    ]
    FUNC_ITEMS = Callable[[Any], Iterable[tuple[Any, Any]]]
    FUNC_CONTAINS = Callable[[Any, Any], bool]
    FUNC_GET = Callable[[Any, Any], Any]
    FUNC_SET = Callable[[Any, Any, Any], None]
    FUNC_CONSTRUCT = Callable[[], Any]
    RECURSIVE_DTYPE_FUNCS = tuple[FUNC_ITEMS, FUNC_CONTAINS, FUNC_GET, FUNC_SET, FUNC_CONSTRUCT]


def recursive_update(
    source: Any,
    addon: Any,
    recursive_types: dict[type | tuple[type, ...], RECURSIVE_DTYPE_FUNCS] | None = None,
    types: dict[type | tuple[type, ...], UPDATE_OPTIONS | tuple[UPDATE_OPTIONS, UPDATE_OPTIONS]] | None = None,
    paths: dict[str, UPDATE_OPTIONS] | None = None,
    constructor: Callable[[], Any] | None = None,
    undefined_new: UPDATE_OPTIONS = "write",
    undefined_existing: UPDATE_OPTIONS = "skip",
    type_mismatch: UPDATE_OPTIONS = "raise",
) -> dict[str, list[str]]:
    """Recursively update a complex data structure using another data structure.

    Parameters
    ----------
    source
        Data structure to update in place.
        The type of this object must be
        defined in the `recursive_types` argument.
    addon
        Data structure containing additional data to update `source` with.
    recursive_types
        Definition of recursive data types.
        Each key is a type (or tuple of types) used
        to identify data types with the `isinstance` function.
        Each value is a tuple of three functions:
        1. Function to extract items from the data. It must accept an instance
           of the type and return an iterable of key-value pairs.
        2. Function to check if a key is in the data. It must accept an instance
           of the type and a key, and return a boolean.
        3. Function to get a value from the data. It must accept an instance
           of the type and a key, and return the value.
        4. Function to set a key-value pair in the data. It must accept an instance
           of the type, a key, and a value, respectively.
        5. Function to construct a new instance of the type. It must accept no arguments.

        By default, `dict` is defined as follows:
        ```python
        recursive_types = {
            dict: (
                lambda dic: dic.items(),
                lambda dic, key: key in dic,
                lambda dic, key: dic[key],
                lambda dic, key, value: dic.update({key: value}),
                lambda: dict(),
            )
        }
        ```
        This default argument will be updated/overwritten with any custom types provided.
        For example, to add support for a custom class `MyClass`, you can use:
        ```python
        recursive_types = {
            MyClass: (
                lambda obj: vars(object).items(),
                lambda obj, key: hasattr(obj, key),
                lambda obj, key: getattr(obj, key),
                lambda obj, key, value: setattr(obj, key, value),
                lambda: MyClass(),
            )
        }
        ```
    types
        Update behavior for specific data types.
        Each key is a type (or tuple of types) used
        to identify data types with the `isinstance` function.
        Each value is can be single update option for all cases,
        or a tuple of two update options for when the corresponding
        key/attribute does not exist in the source data, and when it does, respectively.
        Each update option can be either a keyword specifying a predefined behavior,
        or a function for custom behavior. The available keywords are:
        - "skip": Ignore the key/attribute; do not change anything.
        - "write": Write the key/attribute in source data with the value from the addon, overwriting if it exists.

        A custom function must accept a single argument, a tuple of either one or two values.
        If it is two values, the first value is the current value in the source data,
        and the second value is the value from the addon.
        If it is one value, the value is the value from the addon,
        meaning the key/attribute does not exist in the source data.
        The function must either return `None` (for when nothing must be changes in the source)
        or a tuple of two values:
        1. The new value to write in the source data.
        2. A string specifying the change type.

        By default, the following behavior is defined for basic types:
        ```python
        types = {list: ("write", lambda data: (data[0] + data[1], "append"))}
        ```

        The default behavior for any data type not specified in this argument
        is determined by the `undefined_new` and `undefined_existing` arguments.
    paths
        Update behavior for specific keys using JSONPath expressions.
        This is the same as the `types` argument, but targeting specific keys
        instead of data types.
        Everything is the same as in the `types` argument, except that the keys
        are JSONPath expressions as strings.
    constructor
        Custom constructor for creating new instances of the source data type.
        This is used when the addon data contains a recursive key/attribute
        that is not present in the source data. If not provided,
        the type of the addon value will be used to create a new instance.
    undefined_new
        Behavior for when a non-recursive data with no defined behavior
        in the `types` argument is found in the addon data
        but not in the source data.
    undefined_existing
        Behavior for when a non-recursive data with no defined behavior
        in the `types` argument is found in the addon data
        and in the source data.
    type_mismatch
        Behavior for when a key/attribute in the source data
        is not a recursive type, but the corresponding key/attribute
        in the addon data is a recursive type.
    """
    def get_funcs(data: Any) -> RECURSIVE_DTYPE_FUNCS | None:
        for typ, funcs in recursive_types.items():
            if isinstance(data, typ):
                return funcs
        return None

    def recursive(
        src: Any,
        add: Any,
        src_funcs: RECURSIVE_DTYPE_FUNCS,
        add_funcs: RECURSIVE_DTYPE_FUNCS,
        path: str,
    ):

        def apply(behavior: UPDATE_OPTIONS | tuple[UPDATE_OPTIONS, UPDATE_OPTIONS]):
            action = behavior[int(key_exists_in_src)] if isinstance(behavior, tuple) else behavior
            if action == "raise":
                raise_error(typ="duplicate")
            elif action == "skip":
                change_type = "skip"
            elif action == "write":
                change_type = "write"
                fn_src_set(src, key, value)
            elif not isinstance(action, str):
                out = action((source_value, value) if key_exists_in_src else (value,))
                if out:
                    new_value, change_type = out
                    fn_src_set(src, key, new_value)
                else:
                    change_type = "skip"
            else:
                raise ValueError(f"Invalid update behavior '{action}' for key '{key}' at path '{path}'.")
            log[fullpath] = (type(source_value) if key_exists_in_src else None, type(value), change_type)
            return

        def raise_error(typ: Literal["duplicate", "type_mismatch"]):
            raise _exception.update.PySerialsUpdateRecursiveDataError(
                problem_type=typ,
                path=fullpath,
                data=src[key],
                data_full=src,
                data_addon=value,
                data_addon_full=addon,
            )

        _, fn_src_contains, fn_src_get, fn_src_set, _ = src_funcs
        fn_add_items, _, _, _, fn_add_construct = add_funcs

        for key, value in fn_add_items(add):
            fullpath = f"{path}.'{key}'"
            try:
                full_jpath = _jsonpath.parse(fullpath)  # quote to avoid JSONPath syntax errors
            except Exception as e:
                print(fullpath)
                raise e
            key_exists_in_src = fn_src_contains(src, key)
            source_value = fn_src_get(src, key) if key_exists_in_src else None
            for jpath_str, matches in jsonpath_match.items():
                if full_jpath in matches:
                    apply(paths[jpath_str])
                    break
            else:
                for typ, action in type_to_arg.items():
                    if isinstance(value, typ):
                        apply(action)
                        break
                else:
                    funcs_value = get_funcs(value)
                    if funcs_value:
                        # Value is a recursive type
                        if key_exists_in_src:
                            funcs_src_value = get_funcs(source_value)
                            if not funcs_src_value:
                                # Source value is not a recursive type
                                apply(type_mismatch)
                            else:
                                recursive(
                                    src=fn_src_get(src, key),
                                    add=value,
                                    path=fullpath,
                                    src_funcs=src_funcs,
                                    add_funcs=funcs_value,
                                )
                        else:
                            # Source value does not exist; create a new instance
                            new_instance = constructor() if constructor else fn_add_construct()
                            fn_src_set(src, key, new_instance)
                            funcs_src_value = get_funcs(new_instance)
                            recursive(
                                src=new_instance,
                                add=value,
                                path=fullpath,
                                src_funcs=funcs_src_value,
                                add_funcs=funcs_value,
                            )
                    else:
                        # addon value is of a non-recursive type that does not have any defined behavior;
                        # Apply the default behavior for of ("write", "skip") for the key.
                        apply(undefined_existing if key_exists_in_src else undefined_new)
        return

    type_to_arg = {list: ("write", lambda data: (data[0] + data[1], "append"))} | (types or {})
    recursive_types = {
        dict: (
            lambda dic: dic.items(),
            lambda dic, key: key in dic,
            lambda dic, key: dic[key],
            lambda dic, key, value: dic.update({key: value}),
            lambda: dict(),
        )
    } | (recursive_types or {})
    jsonpath_match = {
        jpath_str: [match.full_path for match in _jsonpath.parse(jpath_str).find(addon)]
        for jpath_str in (paths or {}).keys()
    }
    log = {}
    funcs_src = get_funcs(source)
    funcs_add = get_funcs(addon)
    for funcs, param_name, data in ((funcs_src, "source", source), (funcs_add, "addon", addon)):
        if not funcs:
            raise ValueError(f"Data type '{type(data)}' of '{param_name}' is not provided in 'recursive_types'.")
    recursive(
        src=source,
        add=addon,
        path="$",
        src_funcs=funcs_src,
        add_funcs=funcs_add,
    )
    return log


def data_from_jsonschema(data: dict | list, schema: dict) -> None:
    """Fill missing data in a data structure with default values from a JSON schema."""
    if 'properties' in schema:
        for prop, subschema in schema['properties'].items():
            if 'default' in subschema:
                data.setdefault(prop, subschema['default'])
            if prop in data:
                data_from_jsonschema(data[prop], subschema)
    elif 'items' in schema and isinstance(data, list):
        for item in data:
            data_from_jsonschema(item, schema['items'])
    return


def remove_keys(data: dict | list, keys: str | Sequence[str]):
    def recursive_pop(d):
        if isinstance(d, dict):
            return {k: recursive_pop(v) for k, v in d.items() if k not in keys}
        if isinstance(d, list):
            return [recursive_pop(v) for v in d]
        return d
    if isinstance(keys, str):
        keys = [keys]
    return recursive_pop(data)


class TemplateFiller:

    def __init__(
        self,
        marker_start_value: str = "$",
        marker_end_value: str = "$",
        repeater_start_value: str = "{",
        repeater_end_value: str = "}",
        repeater_count_value: int = 2,
        start_list: str = "$[[",
        start_unpack: str = "*{{",
        start_code: str = "#{{",
        end_list: str = "]]$",
        end_unpack: str = "}}*",
        end_code: str = "}}#",
        raise_no_match: bool = True,
        leave_no_match: bool = False,
        no_match_value: Any = None,
        code_context: dict[str, Any] | None = None,
        code_context_partial: dict[str, Callable | tuple[Callable, str]] | None = None,
        code_context_call: dict[str, Callable[[Callable], Any]] | None = None,
        stringer: Callable[[str], str] = str,
        unpack_string_joiner: str = ", ",
        relative_template_keys: list[str] | None = None,
        relative_key_key: str | None = None,
        implicit_root: bool = True,
        getter_function_name: str = "get",
        skip_key_func: Callable[[list[str]], bool] | None = None,
    ):
        self._marker_start_value = marker_start_value
        self._marker_end_value = marker_end_value
        self._repeater_start_value = repeater_start_value
        self._repeater_end_value = repeater_end_value
        self._repeater_count_value = repeater_count_value
        self._pattern_list = _RegexPattern(start=start_list, end=end_list)
        self._pattern_unpack = _RegexPattern(start=start_unpack, end=end_unpack)
        self._pattern_code = _RegexPattern(start=start_code, end=end_code)

        self._raise_no_match = raise_no_match
        self._leave_no_match = leave_no_match
        self._no_match_value = no_match_value
        self._code_context = code_context or {}
        self._code_context_partial = code_context_partial or {}
        self._code_context_call = code_context_call or {}
        self._stringer = stringer
        self._unpack_string_joiner = unpack_string_joiner
        self._add_prefix = implicit_root
        self._template_keys = relative_template_keys or []
        self._relative_key_key = relative_key_key
        self._getter_function_name = getter_function_name
        self._skip_func = skip_key_func

        self._pattern_value: dict[int, _RegexPattern] = {}
        self._data = None
        self._visited_paths = {}
        return

    def fill(
        self,
        data: dict | list,
        template: dict | list | str | None = None,
        current_path: str = "",
    ):
        self._data = data
        self._visited_paths = {}
        path = _jsonpath.parse((f"$.{current_path}" if self._add_prefix else current_path) if current_path else "$")
        return self._recursive_subst(
            templ=template or data,
            current_path=path,
            relative_path_anchor=path,
            level=0,
            current_chain=(path,),
        )

    def _recursive_subst(self, templ, current_path: str, relative_path_anchor: str, level: int, current_chain: tuple[str, ...], is_key: bool = False):

        def get_code_value(match: _re.Match | str):

            def getter_function(path: str, default: Any = None, search: bool = False):
                value, matched = get_address_value(path, return_all_matches=search, from_code=True)
                if matched:
                    return value
                if search:
                    return []
                return default

            code_str = match if isinstance(match, str) else match.group(1)
            code_lines = ["def __inline_code__():"]
            code_lines.extend([f"    {line}" for line in code_str.strip("\n").splitlines()])
            code_str_full = "\n".join(code_lines)
            global_context = self._code_context.copy() | {
                self._getter_function_name: getter_function,
                "__current_path__": current_path,
                "__relative_path_anchor__": relative_path_anchor
            }
            for name, partial_func_data in self._code_context_partial.items():
                if isinstance(partial_func_data, tuple):
                    func, arg_name = partial_func_data
                    global_context[name] = _partial(func, **{arg_name: getter_function})
                else:
                    global_context[name] = _partial(partial_func_data, getter_function)
            for name, call_func in self._code_context_call.items():
                global_context[name] = call_func(getter_function)
            local_context = {}
            try:
                exec(code_str_full, global_context, local_context)
                output = local_context["__inline_code__"]()
            except Exception as e:
                raise_error(
                    description_template=f"Code at {{path_invalid}} raised an exception: {e}\n{code_str_full}",
                    path_invalid=current_path,
                    exception=e,
                )
            return output

        def get_address_value(match: _re.Match | str, return_all_matches: bool = False, from_code: bool = False):
            raw_path = match if isinstance(match, str) else str(match.group(1))
            path, num_periods = self._remove_leading_periods(raw_path.strip())
            if num_periods == 0:
                path = f"$.{path}" if self._add_prefix else path
            try:
                path_expr = _jsonpath.parse(path)
            except _jsonpath_exceptions.JSONPathError:
                raise_error(
                    path_invalid=path,
                    description_template="JSONPath expression {path_invalid} is invalid.",
                )
            if num_periods:
                if relative_path_anchor != current_path:
                    anchor_path = relative_path_anchor if is_relative_template else current_path
                else:
                    anchor_path = current_path
                root_path_expr = anchor_path
                for period in range(num_periods):
                    if isinstance(root_path_expr, _jsonpath.Root):
                        raise_error(
                            path_invalid=path_expr,
                            description_template=(
                                "Relative path {path_invalid} is invalid; "
                                f"reached root but still {num_periods - period} levels remaining."
                            ),
                        )
                    root_path_expr = root_path_expr.left
                # Handle relative-key key
                if self._relative_key_key and path == self._relative_key_key:
                    output = root_path_expr.right
                    if isinstance(output, _jsonpath.Fields):
                        output = output.fields[0]
                    elif isinstance(output, _jsonpath.Index):
                        output = output.index
                    if from_code:
                        return output, True
                    return output
                path_expr = self._concat_json_paths(root_path_expr, path_expr)
            # print("IN GET ADD VAL", path_expr)
            cached_result = self._visited_paths.get(path_expr)
            if cached_result:
                value, matched = cached_result
            else:
                value, matched = get_value(path_expr, return_all_matches, from_code)
            if not self._is_relative_template(path_expr):
                self._visited_paths[path_expr] = (value, matched)
            if from_code:
                return value, matched
            if matched:
                return value
            if self._leave_no_match:
                return match.group()
            return self._no_match_value

        def get_value(jsonpath, return_all_matches: bool, from_code: bool) -> tuple[Any, bool]:
            matches = recursive_match(jsonpath)
            if not matches:
                if from_code:
                    return None, False
                if return_all_matches:
                    return [], True
                if self._raise_no_match:
                    raise_error(
                        path_invalid=jsonpath,
                        description_template="JSONPath expression {path_invalid} did not match any data.",
                    )
                return None, False
            values = [m.value for m in matches]
            output = values if return_all_matches or len(values) > 1 else values[0]
            if relative_path_anchor == current_path:
                path_fields = self._extract_fields(jsonpath)
                has_template_key = any(field in self._template_keys for field in path_fields)
                _rel_path_anchor = current_path if has_template_key else str(jsonpath)
            else:
                _rel_path_anchor = relative_path_anchor
            return self._recursive_subst(
                output,
                current_path=jsonpath,
                relative_path_anchor=_rel_path_anchor,
                level=0,
                current_chain=current_chain + (jsonpath,),
            ), True

        def recursive_match(expr) -> list:
            matches = expr.find(self._data)
            if matches:
                return matches
            if isinstance(expr.left, _jsonpath.Root):
                return []
            whole_matches = []
            left_matches = recursive_match(expr.left)
            for left_match in left_matches:
                left_match_filled = self._recursive_subst(
                    templ=left_match.value,
                    current_path=expr.left,
                    relative_path_anchor=expr.left,
                    level=0,
                    current_chain=current_chain + (expr.left,),
                ) if isinstance(left_match.value, str) else left_match.value
                right_matches = expr.right.find(left_match_filled)
                whole_matches.extend(right_matches)
            return whole_matches

        def get_relative_path(new_path):
            return new_path if current_path == relative_path_anchor else relative_path_anchor

        def fill_nested_values(match: _re.Match | str):
            pattern_nested = self._get_value_regex_pattern(level=level + 1)
            return pattern_nested.sub(
                lambda x: str(
                    self._recursive_subst(
                        templ=x.group(),
                        current_path=current_path,
                        relative_path_anchor=get_relative_path(current_path),
                        level=level + 1,
                        current_chain=current_chain,
                    )
                ),
                match if isinstance(match, str) else match.group(1),
            )

        def string_filler_unpack(match: _re.Match):
            path = str(match.group(1)).strip()
            match_list = self._pattern_list.fullmatch(path)
            if match_list:
                values = get_address_value(match_list, return_all_matches=True)
            else:
                match_code = self._pattern_code.fullmatch(path)
                if match_code:
                    values = get_code_value(match_code)
                else:
                    values = get_address_value(path)
            return self._unpack_string_joiner.join([self._stringer(val) for val in values])

        def raise_error(path_invalid: str, description_template: str, exception: Exception | None = None):
            raise _exception.update.PySerialsUpdateTemplatedDataError(
                description_template=description_template,
                path_invalid=path_invalid,
                path=current_path,
                data=templ,
                data_full=self._data,
                data_source=self._data,
                template_start=self._marker_start_value,
                template_end=self._marker_end_value,
            ) from exception

        # print("IN MAIN", self._extract_fields(current_path))

        if self._skip_func and self._skip_func(self._extract_fields(current_path)):
            return templ

        if current_path in self._visited_paths:
            return self._visited_paths[current_path][0]

        self._check_endless_loop(templ, current_chain)
        is_relative_template = self._is_relative_template(current_path)

        if isinstance(templ, str):
            # Handle value blocks
            pattern_value = self._get_value_regex_pattern(level=level)
            if match_value := pattern_value.fullmatch(templ):
                out = get_address_value(fill_nested_values(match_value))
            # Handle list blocks
            elif match_list := self._pattern_list.fullmatch(templ):
                out = get_address_value(fill_nested_values(match_list), return_all_matches=True)
            # Handle code blocks
            elif match_code := self._pattern_code.fullmatch(templ):
                out = get_code_value(match_code)
            # Handle unpack blocks
            elif match_unpack := self._pattern_unpack.fullmatch(templ):
                unpack_value = match_unpack.group(1)
                if submatch_code := self._pattern_code.fullmatch(unpack_value):
                    out = get_code_value(submatch_code)
                else:
                    unpack_value = fill_nested_values(unpack_value)
                    if submatch_list := self._pattern_list.fullmatch(unpack_value):
                        out = get_address_value(submatch_list, return_all_matches=True)
                    else:
                        out = get_address_value(unpack_value)
            # Handle strings
            else:
                code_blocks_filled = self._pattern_code.sub(
                    lambda x: self._stringer(get_code_value(x)),
                    templ
                )
                nested_values_filled = fill_nested_values(code_blocks_filled)
                unpacked_filled = self._pattern_unpack.sub(string_filler_unpack, nested_values_filled)
                lists_filled = self._pattern_list.sub(
                    lambda x: self._stringer(get_address_value(x)),
                    unpacked_filled
                )
                out = pattern_value.sub(
                    lambda x: self._stringer(get_address_value(x)),
                    lists_filled
                )
            if not is_relative_template and not is_key:
                self._visited_paths[current_path] = (out, True)
            return out

        if isinstance(templ, list):
            out = []
            for idx, elem in enumerate(templ):
                new_path = _jsonpath.Child(current_path, _jsonpath.Index(idx))
                elem_filled = self._recursive_subst(
                    templ=elem,
                    current_path=new_path,
                    relative_path_anchor=get_relative_path(new_path),
                    level=0,
                    current_chain=current_chain + (new_path,),
                )
                if isinstance(elem, str) and self._pattern_unpack.fullmatch(elem):
                    try:
                        out.extend(elem_filled)
                    except TypeError as e:
                        raise_error(
                            path_invalid=current_path,
                            description_template=str(e)
                        )
                else:
                    out.append(elem_filled)
            if not is_relative_template:
                self._visited_paths[current_path] = (out, True)
            return out

        if isinstance(templ, dict):
            new_dict = {}
            addons = []
            for key, val in templ.items():
                key_filled = self._recursive_subst(
                    templ=key,
                    current_path=current_path,
                    relative_path_anchor=relative_path_anchor,
                    level=0,
                    current_chain=current_chain,
                    is_key=True,
                )
                if isinstance(key, str) and self._pattern_unpack.fullmatch(key):
                    addons.append((key_filled, val))
                    continue
                if key_filled in self._template_keys:
                    new_dict[key_filled] = val
                    continue
                new_path = _jsonpath.Child(current_path, _jsonpath.Fields(key_filled))
                new_dict[key_filled] = self._recursive_subst(
                    templ=val,
                    current_path=new_path,
                    relative_path_anchor=get_relative_path(new_path),
                    level=0,
                    current_chain=current_chain + (new_path,),
                )
            for addon_dict, addon_settings in sorted(
                addons, key=lambda addon: addon[1].get("priority", 0) if addon[1] else 0
            ):
                addon_settings = {k: v for k, v in (addon_settings or {}).items() if k not in ("priority",)}
                recursive_update(
                    source=new_dict,
                    addon=addon_dict,
                    **addon_settings,
                )
            if not is_relative_template:
                self._visited_paths[current_path] = (new_dict, True)
            return new_dict
        return templ

    def _check_endless_loop(self,templ, chain: tuple[str, ...]):
        last_idx = len(chain) - 1
        first_idx = chain.index(chain[-1])
        if first_idx == last_idx:
            return
        loop = [chain[-2], *chain[first_idx: -2]]
        loop_str = "\n".join([f"- {path}" for path in loop])
        history_str = "\n".join([f"- {path}" for path in chain])
        raise _exception.update.PySerialsUpdateTemplatedDataError(
            description_template=f"Path {{path_invalid}} starts a loop:\n{loop_str}\nHistory:\n{history_str}",
            path_invalid=chain[-2],
            path=chain[0],
            data=templ,
            data_full=self._data,
            data_source=self._data,
            template_start=self._marker_start_value,
            template_end=self._marker_end_value,
        )

    def _get_value_regex_pattern(self, level: int = 0) -> _RegexPattern:
        if level in self._pattern_value:
            return self._pattern_value[level]
        count = self._repeater_count_value + level
        pattern = _RegexPattern(
            start=f"{self._marker_start_value}{self._repeater_start_value * count} ",
            end=f" {self._repeater_end_value * count}{self._marker_end_value}",
        )
        self._pattern_value[level] = pattern
        return pattern

    def _is_relative_template(self, jsonpath):
        path_fields = self._extract_fields(jsonpath)
        return any(field in self._template_keys for field in path_fields)

    @staticmethod
    def _remove_leading_periods(s: str) -> (str, int):
        match = _re.match(r"^(\.*)(.*)", s)
        if match:
            leading_periods = match.group(1)
            rest_of_string = match.group(2)
            num_periods = len(leading_periods)
        else:
            num_periods = 0
            rest_of_string = s
        return rest_of_string, num_periods

    @staticmethod
    def _extract_fields(jsonpath):
        def _recursive_extract(expr):
            if hasattr(expr, "fields"):
                fields.extend(expr.fields)
            if hasattr(expr, "right"):
                _recursive_extract(expr.right)
            if hasattr(expr, "left"):
                _recursive_extract(expr.left)
            return
        fields = []
        _recursive_extract(jsonpath)
        return fields

    @staticmethod
    def _normalize_path(path: str) -> str:
        return path.replace("'", "")

    def _concat_json_paths(self, path1, path2):
        if not isinstance(path2, _jsonpath.Child):
            return _jsonpath.Child(path1, path2)
        return _jsonpath.Child(self._concat_json_paths(path1, path2.left), path2.right)

class _RegexPattern:

    def __init__(self, start: str, end: str):
        start_esc = _re.escape(start)
        end_esc = _re.escape(end)
        self.pattern = _re.compile(rf"{start_esc}(.*?)(?={end_esc}){end_esc}", _re.DOTALL)
        return

    def fullmatch(self, string: str) -> _re.Match | None:
        # Use findall to count occurrences of segments in the text
        matches = self.pattern.findall(string)
        if len(matches) == 1:
            # Verify the match spans the entire string
            return self.pattern.fullmatch(string.strip())
        return None

    def sub(self, repl, string: str) -> str:
        return self.pattern.sub(repl, string)




from __future__ import annotations as _annotations
from typing import Literal as _Literal, TYPE_CHECKING as _TYPE_CHECKING
from pathlib import Path as _Path
import json as _json
import ruamel.yaml as _yaml
from ruamel.yaml import scalarstring as _yaml_scalar_string
import tomlkit as _tomlkit

if _TYPE_CHECKING:
    from typing import Callable, Any


def to_string(
    data: dict | list | str | int | float | bool | _yaml.CommentedMap | _yaml.CommentedSeq,
    data_type: _Literal["json", "yaml", "toml"],
    sort_keys: bool = False,
    indent: int | None = None,
    default: Callable[[Any], Any] | None = None,
    end_of_file_newline: bool = True,
    indent_mapping: int = 2,
    indent_sequence: int = 4,
    indent_sequence_offset: int = 2,
    multiline_string_to_block: bool = True,
    remove_top_level_indent: bool = True
):
    if data_type == "json":
        return to_json_string(data, sort_keys=sort_keys, indent=indent, default=default, end_of_file_newline=end_of_file_newline)
    if data_type == "yaml":
        return to_yaml_string(
            data,
            end_of_file_newline=end_of_file_newline,
            indent_mapping=indent_mapping,
            indent_sequence=indent_sequence,
            indent_sequence_offset=indent_sequence_offset,
            multiline_string_to_block=multiline_string_to_block,
            remove_top_level_indent=remove_top_level_indent
        )
    return to_toml_string(data, sort_keys=sort_keys, end_of_file_newline=end_of_file_newline)


def to_yaml_string(
    data: dict | list | str | int | float | bool | _yaml.CommentedMap | _yaml.CommentedSeq,
    end_of_file_newline: bool = True,
    indent_mapping: int = 2,
    indent_sequence: int = 4,
    indent_sequence_offset: int = 2,
    multiline_string_to_block: bool = True,
    remove_top_level_indent: bool = True
) -> str:
    yaml = _yaml.YAML(typ=["rt", "string"])
    yaml.indent(mapping=indent_mapping, sequence=indent_sequence, offset=indent_sequence_offset)
    if multiline_string_to_block:
        _yaml_scalar_string.walk_tree(data)
    yaml_syntax = yaml.dumps(data, add_final_eol=False).removesuffix("\n...")
    if remove_top_level_indent:
        yaml_lines = yaml_syntax.splitlines()
        count_leading_spaces = len(yaml_lines[0]) - len(yaml_lines[0].lstrip(" "))
        if count_leading_spaces:
            yaml_syntax = "\n".join(yaml_line.removeprefix(" " * count_leading_spaces) for yaml_line in yaml_lines)
    return f"{yaml_syntax}\n" if end_of_file_newline else yaml_syntax


def to_toml_string(
    data: dict | list | str | int | float | bool | _yaml.CommentedMap | _yaml.CommentedSeq,
    sort_keys: bool = False,
    end_of_file_newline: bool = True,
) -> str:
    string = _tomlkit.dumps(data, sort_keys=sort_keys)
    return f"{string.rstrip("\n")}\n" if end_of_file_newline else string


def to_json_string(
    data: dict | list | str | int | float | bool | _yaml.CommentedMap | _yaml.CommentedSeq,
    sort_keys: bool = False,
    indent: int | None = None,
    default: Callable[[Any], Any] | None = None,
    end_of_file_newline: bool = True,
) -> str:
    string = _json.dumps(data, indent=indent, sort_keys=sort_keys, default=default)
    return f"{string.rstrip("\n")}\n" if end_of_file_newline else string


def to_json_file(
    data: dict | list | str | int | float | bool | _yaml.CommentedMap | _yaml.CommentedSeq,
    path: str | _Path,
    sort_keys: bool = False,
    indent: int | None = None,
    default: Callable[[Any], Any] | None = None,
    end_of_file_newline: bool = True,
    make_dirs: bool = True,
) -> None:
    json_string = to_json_string(
        data,
        sort_keys=sort_keys,
        indent=indent,
        default=default,
        end_of_file_newline=end_of_file_newline,
    )
    path = _Path(path).resolve()
    if make_dirs:
        path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json_string)
    return


def to_yaml_file(
    data: dict | list | str | int | float | bool | _yaml.CommentedMap | _yaml.CommentedSeq,
    path: str | _Path,
    make_dirs: bool = True,
    indent_mapping: int = 2,
    indent_sequence: int = 4,
    indent_sequence_offset: int = 2,
    multiline_string_to_block: bool = True,
):
    path = _Path(path).resolve()
    if make_dirs:
        path.parent.mkdir(parents=True, exist_ok=True)
    yaml = _yaml.YAML()
    yaml.indent(mapping=indent_mapping, sequence=indent_sequence, offset=indent_sequence_offset)
    if multiline_string_to_block:
        _yaml_scalar_string.walk_tree(data)
    yaml.dump(data, path)
    return



from typing import Literal as _Literal, Callable as _Callable, Any as _Any
import json as _json
from pathlib import Path as _Path
from functools import partial as _partial
import tomlkit as _tomlkit
from tomlkit.exceptions import TOMLKitError as _TOMLKitError
import ruamel.yaml as _yaml

import pyserials.exception as _exception


def from_file(
    path: str | _Path,
    data_type: _Literal["json", "yaml", "toml"] | None = None,
    json_strict: bool = True,
    yaml_safe: bool = True,
    toml_as_dict: bool = False,
):
    if data_type is None:
        data_type = _Path(path).suffix.removeprefix(".").lower()
        if data_type == "yml":
            data_type = "yaml"
    if data_type not in ("json", "yaml", "toml"):
        raise _exception.read.PySerialsInvalidFileExtensionError(filepath=path)
    if data_type == "json":
        return json_from_file(path=path, strict=json_strict)
    if data_type == "yaml":
        return yaml_from_file(path=path, safe=yaml_safe)
    return toml_from_file(path=path, as_dict=toml_as_dict)


def from_string(
    data: str,
    data_type: _Literal["json", "yaml", "toml"],
    json_strict: bool = True,
    yaml_safe: bool = True,
    toml_as_dict: bool = False,
):
    if data_type == "json":
        return json_from_string(data=data, strict=json_strict)
    if data_type == "yaml":
        return yaml_from_string(data=data, safe=yaml_safe)
    return toml_from_string(data=data, as_dict=toml_as_dict)


def toml(
    source: str | _Path, as_dict: bool = False
) -> _tomlkit.TOMLDocument | dict:
    if isinstance(source, str):
        return toml_from_string(data=source, as_dict=as_dict)
    return toml_from_file(path=source, as_dict=as_dict)


def toml_from_string(data: str, as_dict: bool = False) -> _tomlkit.TOMLDocument | dict:
    content = _read_from_string(data=data, data_type="toml", loader=_tomlkit.loads, exception=_TOMLKitError)
    if as_dict:
        return dict(content)
    return content


def toml_from_file(path: str | _Path, as_dict: bool = False) -> _tomlkit.TOMLDocument | dict:
    content = _read_from_file(
        path=path,
        data_type="toml",
        loader=_tomlkit.loads,
        exception=_TOMLKitError
    )
    if as_dict:
        return dict(content)
    return content


def json(source: str | _Path, strict: bool = True) -> dict | list | str | int | float | bool:
    if isinstance(source, str):
        return json_from_string(data=source, strict=strict)
    return json_from_file(path=source, strict=strict)


def json_from_string(data: str, strict: bool = True) -> dict | list | str | int | float | bool:
    content = _read_from_string(
        data=data,
        data_type="json",
        loader=_partial(_json.loads, strict=strict),
        exception=_json.JSONDecodeError
    )
    return content


def json_from_file(path: str | _Path, strict: bool = True) -> dict | list | str | int | float | bool:
    content = _read_from_file(
        path=path,
        data_type="json",
        loader=_partial(_json.loads, strict=strict),
        exception=_json.JSONDecodeError
    )
    return content


def yaml(
    source: str | _Path,
    safe: bool = True,
    constructors: dict[str, _Callable[[_yaml.Constructor, _yaml.ScalarNode], _Any]] | None = None,
) -> dict | list | str | int | float | bool | _yaml.CommentedMap | _yaml.CommentedSeq:
    """Load YAML data from a file or string.

    Parameters
    ----------
    source : str | _Path
        Path to the file or YAML data string.
    safe : bool, default: True
        Use safe YAML loader.
    constructors : dict[str, Callable], default: None
        Custom YAML constructors.

    References
    ----------
    - See `add_constructor` method in
      `ruamel.yaml.constructor.SafeConstructor` and `ruamel.yaml.constructor.RoundTripConstructor`.
    - See `add_constructor` in [PyYAML documentation](https://pyyaml.org/wiki/PyYAMLDocumentation).
    - See Tags in [YAML documentation](https://yaml.org/spec/1.2.0/#id2560445)
    """
    if isinstance(source, str):
        return yaml_from_string(data=source, safe=safe, constructors=constructors)
    return yaml_from_file(path=source, safe=safe)


def yaml_from_string(
    data: str,
    safe: bool = True,
    constructors: dict[str, _Callable[[_yaml.Constructor, _yaml.ScalarNode], _Any]] | None = None,
) -> dict | list | str | int | float | bool | _yaml.CommentedMap | _yaml.CommentedSeq:
    content = _read_from_string(
        data=data,
        data_type="yaml",
        loader=_make_yaml_loader(safe=safe, constructors=constructors).load,
        exception=_yaml.YAMLError
    )
    return content


def yaml_from_file(
    path: str | _Path,
    safe: bool = True,
    constructors: dict[str, _Callable[[_yaml.Constructor, _yaml.ScalarNode], _Any]] | None = None,
) -> dict | list | str | int | float | bool | _yaml.CommentedMap | _yaml.CommentedSeq:
    content = _read_from_file(
        path=path,
        data_type="yaml",
        loader=_make_yaml_loader(safe=safe, constructors=constructors).load,
        exception=_yaml.YAMLError
    )
    return content


def _read_from_file(
    path: str | _Path,
    data_type: _Literal["json", "yaml", "toml"],
    loader,
    exception,
):
    path = _Path(path).resolve()
    if not path.is_file():
        raise _exception.read.PySerialsMissingFileError(data_type=data_type, filepath=path)
    data = path.read_text()
    if data.strip() == "":
        raise _exception.read.PySerialsEmptyFileError(data_type=data_type, filepath=path)
    try:
        content = loader(data)
    except exception as e:
        raise _exception.read.PySerialsInvalidDataError(
            source_type="file", filepath=path, data_type=data_type, data=data, cause=e
        ) from None
    return content


def _read_from_string(
    data: str,
    data_type: _Literal["json", "yaml", "toml"],
    loader,
    exception,
):
    if not data.strip():
        raise _exception.read.PySerialsEmptyStringError(data_type=data_type)
    try:
        content = loader(data)
    except exception as e:
        raise _exception.read.PySerialsInvalidDataError(
            source_type="string", data_type=data_type, data=data, cause=e
        ) from None
    return content


def _make_yaml_loader(
    safe: bool,
    constructors: dict[str, _Callable[[_yaml.Constructor, _yaml.ScalarNode], _Any]] | None = None
) -> _yaml.YAML:
    loader = _yaml.YAML(typ="safe" if safe else "rt")
    if constructors:
        for key, constructor in constructors.items():
            loader.constructor.add_constructor(key, constructor)
    return loader



import copy as _copy


class PropertyDict:

    def __init__(self, data: dict | None):
        self._data = data or {}
        return

    @property
    def as_dict(self) -> dict:
        """The data as a dictionary."""
        return self._data

    def items(self):
        return self._data.items()

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()
    
    def get(self, key, default=None):
        return self._data.get(key, default)

    def pop(self, key, default=None):
        return self._data.pop(key, default)
    
    def update(self, *args, **kwargs):
        return self._data.update(*args, **kwargs)

    def setdefault(self, key, default):
        return self._data.setdefault(key, default)

    def __getattr__(self, name):
        try:
            return self._data[name]
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __contains__(self, key):
        return key in self._data

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return f"PropertyDict({self._data})"

    def __str__(self):
        return str(self._data)

    def __eq__(self, other):
        return self._data == other._data

    def __ne__(self, other):
        return self._data != other._data

    def __deepcopy__(self, memo):
        # Dynamically handle subclass instantiation with additional arguments
        cls = type(self)
        # Deepcopy the data dictionary
        copied_data = _copy.deepcopy(self._data, memo)
        # Capture extra attributes (everything except `_data`)
        extra_attrs = {
            key: _copy.deepcopy(value, memo)
            for key, value in self.__dict__.items()
            if key != "_data"
        }
        # Create a new instance
        obj = cls.__new__(cls)  # Avoid calling __init__ directly
        obj._data = copied_data  # Set _data directly
        # Restore extra attributes
        obj.__dict__.update(extra_attrs)
        return obj



def items(source, target, path: str = "$"):
    def recursive_compare(src, trg, curr_path):
        if type(src) is not type(trg):
            comp["modified"].append(curr_path)
            return
        if isinstance(src, dict):
            for key in src:
                if key not in trg:
                    comp["added"].append(f"{curr_path}.{key}")
                    continue
                recursive_compare(src[key], trg[key], f"{curr_path}.{key}")
            for key in trg:
                if key not in src:
                    comp["removed"].append(f"{curr_path}.{key}")
            return
        if isinstance(src, (list, tuple)):
            len_src = len(src)
            len_trg = len(trg)
            min_len = min(len_src, len_trg)
            for i in range(min_len):
                recursive_compare(src[i], trg[i], f"{curr_path}[{i}]")
            for i in range(min_len, max(len_src, len_trg)):
                comp["added" if len_src > len_trg else "removed"].append(f"{curr_path}[{i}]")
            return
        comp["unchanged" if src == trg else "modified"].append(curr_path)
        return
    comp = {"added": [], "removed": [], "modified": [], "unchanged": []}
    recursive_compare(source, target, path)
    return {key: sorted(comp[key]) for key in comp}
